"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const graphql = __importStar(require("graphql"));
function mkdesc(desc) {
    if (!desc)
        return '';
    return `/** ${desc} */\n`;
}
function parseEnum(type, enumType) {
    if (enumType == 'type') {
        return `${mkdesc(type.description)}export type ${type.name} = ${type.getValues().map(value => `'${value.name}'`).join(' | ')}`;
    }
    else {
        const vals = type.getValues().map(value => `${mkdesc(value.description)}${value.name} = '${value.name}'`).join(',\n').replace(/\n/g, '\n\t');
        return `${mkdesc(type.description)}export const enum ${type.name} {\n\t${vals}\n}`;
    }
}
//首字母大写
function upperCaseName(name) {
    return name[0].toUpperCase() + name.substr(1);
}
//基础类型
function baseTypes(custom) {
    const gqlKeys = ['String', 'Boolean', 'Int', 'Float', 'ID'];
    //处理类型
    if (!custom.String)
        custom.String = 'string';
    if (!custom.Boolean)
        custom.Boolean = 'boolean';
    if (!custom.Int)
        custom.Int = 'number';
    if (!custom.Float)
        custom.Float = 'number';
    if (!custom.ID)
        custom.ID = 'string | number';
    //返回类型
    return [
        ...Object.keys(custom)
            .map(key => {
            if (gqlKeys.indexOf(key) < 0)
                return;
            return `export type ${key} = ${custom[key]}\n`;
        })
            .filter(s => !!s),
        //函数定义
        'interface GQLFunctionBase<P, R> {',
        '	(a: P): R',
        '	args: P',
        '}',
        '',
        'type GQLFunction<P, R> = GQLFunctionBase<P, R> & R',
    ].join('\n');
}
function parseScalar(type, customscalarTypes) {
    const typeName = (customscalarTypes || {})[type.name] || 'any';
    return `${mkdesc(type.description)}export type ${type.name} = ${typeName}`;
}
function type2ts(type, option) {
    const { nullableType, defaultValue, isNotNull, isArgument } = option;
    //此函数用于自动加入空处理
    const addNull = (val) => (isNotNull || !nullableType) ? val : nullableType({ type: val, defaultVal: defaultValue, isArgument });
    //分别处理不同类型
    if (type.kind == 'ListType')
        return addNull(`Array<${type2ts(type.type, { nullableType, defaultValue: null, isNotNull: false, isArgument: false })}>`);
    else if (type.kind == 'NonNullType')
        return `${type2ts(type.type, { nullableType, defaultValue: null, isNotNull: true, isArgument: false })}`;
    else
        return addNull(type.name.value);
}
function parseObject(type, objectType, argument2interface, nullableType) {
    const argTypes = [];
    const fields = type.getFields();
    const types = Object.keys(fields).map(key => {
        const desc = fields[key].description || undefined;
        const ast = fields[key].astNode;
        if (!ast)
            return null;
        //如果有参数则返回函数
        if ((ast.kind == 'FieldDefinition') && ast.arguments && ast.arguments.length) {
            const args = ast.arguments.map(arg => {
                return `${arg.name.value}: ${type2ts(arg.type, { nullableType, defaultValue: arg.defaultValue || null, isNotNull: false, isArgument: true })}`;
            });
            if (argument2interface) {
                const interfaceName = `I${upperCaseName(key)}On${upperCaseName(type.name)}Arguments`;
                argTypes.push(`interface ${interfaceName} {\n\t${args.join('\n\t')}\n}`);
                return `${mkdesc(desc)}${key}: GQLFunction<${interfaceName}, ${type2ts(ast.type, { nullableType, defaultValue: null, isNotNull: false, isArgument: false })}>`;
            }
            else {
                return `${mkdesc(desc)}${key}: GQLFunction<${args.join(', ')} }, ${type2ts(ast.type, { nullableType, defaultValue: null, isNotNull: false, isArgument: false })}>`;
            }
        }
        //否则返回类型
        else
            return `${mkdesc(desc)}${key}: ${type2ts(ast.type, { nullableType, defaultValue: null, isNotNull: false, isArgument: false })}`;
    }).filter(t => !!t).map(type => (type || '').replace(/\n/g, '\n\t'));
    //加入节点类型
    if (objectType == 'interface')
        argTypes.push(`${mkdesc(type.description || undefined)}export interface ${type.name} {\n\t${types.join('\n\t')}\n}`);
    else
        argTypes.push(`${mkdesc(type.description || undefined)}export type ${type.name} = {\n\t${types.join(',\n\t')}\n}`);
    //返回数据
    return argTypes;
}
function parseUnion(type) {
    return `${mkdesc(type.description)}export type ${type.name} = ${type.getTypes().map(type => type.name).join(' | ')}`;
}
/**
 * 转换一个graphql schema为typescript定义文件
 * @param schema 要转换的schema
 * @param param1 转换选项
 */
function parse(schema, { enumType = 'type', objectType = 'interface', namespaceName = 'gql', outputType = 'declare', argument2interface = true, customscalarTypes = {}, nullableType = {}, notNullWhenDefaultValue = true, } = {}) {
    const defaultNullableType = ({ defaultVal, type }) => {
        if (defaultVal) {
            if (defaultVal.kind != 'NullValue' && notNullWhenDefaultValue)
                return type;
            return `${type} | null`;
        }
        return `${type} | null | undefined`;
    };
    const typeMap = schema.getTypeMap();
    const types = [baseTypes(customscalarTypes)];
    //处理数据类型
    Object.keys(typeMap)
        .map(key => typeMap[key])
        .filter(type => type.astNode || ['Query', 'Mutation'].includes(type.name))
        .map(type => {
        //枚举
        if (type instanceof graphql.GraphQLEnumType)
            return parseEnum(type, enumType);
        //输入类型定义
        else if (type instanceof graphql.GraphQLInputObjectType)
            return parseObject(type, objectType, argument2interface, nullableType.input || defaultNullableType);
        //接口定义
        else if (type instanceof graphql.GraphQLInterfaceType)
            return parseObject(type, objectType, argument2interface, nullableType.interface || defaultNullableType);
        //类型定义
        else if (type instanceof graphql.GraphQLObjectType)
            return parseObject(type, objectType, argument2interface, nullableType.object || defaultNullableType);
        //scalar类型定义，这里不处理，因为在baseTypes中已经处理
        else if (type instanceof graphql.GraphQLScalarType)
            return parseScalar(type, customscalarTypes);
        //联合类型定义
        else if (type instanceof graphql.GraphQLUnionType)
            return parseUnion(type);
    })
        .map(type => {
        if (!type)
            return;
        if (type instanceof Array)
            types.push(...type);
        else
            types.push(type);
    });
    //生成结果
    return `${(outputType == 'none') ? '' : `${outputType} `}namespace ${namespaceName} {\n\n\t${types.map(type => type.replace(/\r?\n/g, '\n\t')).join('\n\n\t')}\n\n}`;
}
exports.parse = parse;
